//height map rendering

#include <stdlib.h> //for exit call
#include <GL/glut.h>
#include <math.h> //for sin
#include <iostream> //for FILE

#define ESCAPE 27 //define ASCII value for special keystrokes.

/* easy indexing for texture array */
#define CEMENT 0
#define ROCK 1
#define WATER 2

/* structure for the storing the image data */
struct Image
{
    unsigned long size_x;
    unsigned long size_y;
    char *data;
};

typedef struct Image Image;

const int g_textureCount = 3; //specifies # of textures

Image g_myTextureData[g_textureCount]; //array storing image texture info
GLuint g_theTexture[g_textureCount]; //array storing OpenGL texture info
GLint g_windowWidth = 640, window_height = 480; //window dimensions

double theta = 45.0; //defined for the rotation of our entire scene
const GLint g_mapDimension = 16; //dimension of our hightmap

/* the heightmap used for the rendering of the qireframe */
GLfloat wireframeHeightMap[g_mapDimension][g_mapDimension];
/* the heightmap used for the rendering of the ground */
GLfloat surfaceHeightMap[g_mapDimension][g_mapDimension];

/* used for animation calculations */
GLfloat temp_map[g_mapDimension][g_mapDimension];

int duration = 0;
int duration_counter = 0;

/* texture filename list */
char* texture_filenames[g_textureCount] = {"cement.bmp",
										  "rock.bmp",
										  "water.bmp"};

/*
=================
imageLoaderer
-loads our bitmap data into the Image struct
=================
*/
int imageLoader(const char *filename, Image *image)
{
    FILE *file;

    unsigned long size;
    unsigned long i;
    unsigned short int planes;
    unsigned short int bpp;

    char temp;
	char finalName[80];

	glTexCoord2f(1.0, 0.0);

	strcpy(finalName, "" );
	strcat(finalName, filename);

    if ((file = fopen(finalName, "rb"))==NULL) 
	{
		printf("File Not Found : %s\n",finalName);
		return 0;
    }

    fseek(file, 18, SEEK_CUR);

	glTexCoord2f(1.0, 0.0);
    
	if ((i = fread(&image->size_x, 4, 1, file)) != 1) 
	{
		printf("Error reading width from %s.\n", finalName);
		return 0;
    }

    if ((i = fread(&image->size_y, 4, 1, file)) != 1) 
	{
		printf("Error reading height from %s.\n", finalName);
		return 0;
    }

    size = image->size_x * image->size_y * 3;

    if ((fread(&planes, 2, 1, file)) != 1) 
	{
		printf("Error reading planes from %s.\n", finalName);
		return 0;
    }

    if (planes != 1) 
	{
		printf("Planes from %s is not 1: %u\n", finalName, planes);
		return 0;
    }

    if ((i = fread(&bpp, 2, 1, file)) != 1) 
	{
		printf("Error reading bpp from %s.\n", finalName);
		return 0;
    }

    if (bpp != 24) 
	{
		printf("Bpp from %s is not 24: %u\n", finalName, bpp);
		return 0;
    }

    fseek(file, 24, SEEK_CUR);

    image->data = (char *) malloc(size);

    if (image->data == NULL) 
	{
		printf("Error allocating memory for color-corrected image data");
		return 0;
    }

    if ((i = fread(image->data, size, 1, file)) != 1) 
	{
		printf("Error reading image data from %s.\n", finalName);
		return 0;
    }

    for (i=0;i<size;i+=3) 
	{
		temp = image->data[i];
		image->data[i] = image->data[i+2];
		image->data[i+2] = temp;
    }
    return 1;
}

/*
=================
textureLoader
-physically loads the BMPs and converts them to texture data
=================
*/
void textureLoader()
{
	/* sets the pixel storage, GL_UNPACK_ALIGNMENT : specifies alignment requirements
	   for the start of each pixel row in memory 1 = byte-alignment*/
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	for(int k=0; k < g_textureCount; k++) 
	{
		if(!imageLoader(texture_filenames[k], &g_myTextureData[k])) 
			exit(1);
		
		/* generates texture names */
		glGenTextures(1, &g_theTexture[k]);
		/* creates a named texture bound to a texture target */
	    glBindTexture(GL_TEXTURE_2D, g_theTexture[k]);

		/* sets the texture parameters */
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
		/* builds and loads a set of mipmaps*/
	   	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, g_myTextureData[k].size_x, g_myTextureData[k].size_y, GL_RGB, GL_UNSIGNED_BYTE, g_myTextureData[k].data);
	}
}

/*
=================
switchToOrth
-switches from perspective to orthographic projection
NOTE: (0,0) is situated at the top left corner of your screen
=================
*/
void switchToOrtho() 
{
	/* specifies the Projection matrix as the current matrix */
	glMatrixMode(GL_PROJECTION);
	/* pushes current matrix stack down by one */
	glPushMatrix();
	/* move to the center of the screen */
	glLoadIdentity();
	/* define a 2-D orthographic projection matrix */
	gluOrtho2D(0, g_windowWidth, 0, window_height);
	/* scales along the x, -y, and z axes */
	glScalef(1, -1, 1);
	/* moves along the x, y and z-axis, moves -'window_hight' units (down) on y-axis */
	glTranslatef(0, -1.0*window_height, 0);
	/* specifies the Modelview matrix as the current matrix */
	glMatrixMode(GL_MODELVIEW);
	/* move to the center of the screen */
	glLoadIdentity();
}

/*
=================
switchFromOrth
-switches back from orthographic projection to perspective projection
=================
*/
void switchFromOrtho() 
{
	/* specifies the Projection matrix as the current matrix */
	glMatrixMode(GL_PROJECTION);
	/* pops current matrix stack */
	glPopMatrix();
	/* specifies the Modelview matrix as the current matrix */
	glMatrixMode(GL_MODELVIEW);
}

/*
=================
display
-renders output/scene to the screen
=================
*/
void display(void)
{
	/* set the clear colour - red, green, blue, alpha*/
	glClearColor(1.0, 1.0, 1.0, 0.0);

	/* clear frame buffer and z-buffer */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* swith to orthographic projection, we are currently in perspective projection,
	   the text has to be rendered in orthographic projection mode */
	switchToOrtho();

	/* set text colour to white */
   	glColor3f(1.0,1.0,1.0);
	/* sets the font position roughly in the middle of the window */
	glRasterPos2i(g_windowWidth/2 - 50, 20);

	/* switch back to perspective projection */
	switchFromOrtho();

	/* saves the current state */
	glPushMatrix();
		/* positions the surface in front of the default camera */
		glTranslatef(0.0, 0.0, -3.5);

		/* initial rotation corresponds to isometric drawing angle (30 degrees around x-axis) */
		glRotatef(30.0, 1.0, 0.0, 0.0);
		/***rotation change on keyboard input - theta globally changed***/
		/* now rotate the cam upward (y-axis) at a theta degrees */
		glRotatef(theta, 0.0, 1.0,0.0);

		/*----------
		draw the heightfields
		----------*/

		double mapWidth = 2.0;
		double mapLength = 2.0;
		double mapWidthStep = mapWidth / g_mapDimension;
		double mapLengthStep = mapLength / g_mapDimension;
		double mapStartXPos = 0 - (mapWidth/2.0);
		double mapStartZPos = 0 - (mapLength/2.0);
		double currentXPos = 0.0;
		double currentZPos = 0.0;

		currentZPos = mapStartZPos;
		currentXPos = mapStartXPos;

		for (int i = 0; i < g_mapDimension - 1; i++)
		{
			currentZPos = mapStartZPos;

			for (int j = 0; j < g_mapDimension - 1; j++)
			{
				/* bind texture */
				glBindTexture(GL_TEXTURE_2D, g_theTexture[WATER]);

				/*
				The fluid surface
				*/
				glBegin(GL_LINE_LOOP);
					glTexCoord2f((i + 0.0)/g_mapDimension, (j + 0.0)/g_mapDimension);
					glVertex3f(currentXPos, wireframeHeightMap[i][j], currentZPos);

					glTexCoord2f((i + 1.0)/g_mapDimension, (j + 0.0)/g_mapDimension);
					glVertex3f(currentXPos + mapWidthStep, wireframeHeightMap[i+1][j], currentZPos);

					glTexCoord2f((i + 0.0)/g_mapDimension, (j + 1.0)/g_mapDimension);
					glVertex3f(currentXPos, wireframeHeightMap[i][j+1], currentZPos + mapLengthStep);

					glTexCoord2f((i + 1.0)/g_mapDimension, (j + 0.0)/g_mapDimension);
					glVertex3f(currentXPos + mapWidthStep, wireframeHeightMap[i+1][j], currentZPos);

					glTexCoord2f((i + 1.0)/g_mapDimension, (j + 1.0)/g_mapDimension);
					glVertex3f(currentXPos + mapWidthStep, wireframeHeightMap[i+1][j+1], currentZPos + mapLengthStep);

					glTexCoord2f((i + 0.0)/g_mapDimension, (j + 1.0)/g_mapDimension);
					glVertex3f(currentXPos, wireframeHeightMap[i][j+1], currentZPos + mapLengthStep);
				glEnd();

				glBindTexture(GL_TEXTURE_2D, g_theTexture[ROCK]);

				/*
				The ground/rock surface
				*/
				glBegin(GL_TRIANGLES);
					glTexCoord2f((i + 0.0)/g_mapDimension, (j + 0.0)/g_mapDimension);
					glVertex3f(currentXPos, surfaceHeightMap[i][j], currentZPos);

					glTexCoord2f((i + 1.0)/g_mapDimension, (j + 0.0)/g_mapDimension);
					glVertex3f(currentXPos + mapWidthStep, surfaceHeightMap[i+1][j], currentZPos);

					glTexCoord2f((i + 0.0)/g_mapDimension, (j + 1.0)/g_mapDimension);
					glVertex3f(currentXPos, surfaceHeightMap[i][j+1], currentZPos + mapLengthStep);

					glTexCoord2f((i + 1.0)/g_mapDimension, (j + 0.0)/g_mapDimension);
					glVertex3f(currentXPos + mapWidthStep, surfaceHeightMap[i+1][j], currentZPos);

					glTexCoord2f((i + 1.0)/g_mapDimension, (j + 1.0)/g_mapDimension);
					glVertex3f(currentXPos + mapWidthStep, surfaceHeightMap[i+1][j+1], currentZPos + mapLengthStep);

					glTexCoord2f((i + 0.0)/g_mapDimension, (j + 1.0)/g_mapDimension);
					glVertex3f(currentXPos, surfaceHeightMap[i][j+1], currentZPos + mapLengthStep);
				glEnd();

				currentZPos += mapLengthStep;
			}
			currentXPos += mapWidthStep;
		}

	/* restores previously saved state */
	glPopMatrix();
	
	/* make sure all previously issued commands are completed */
	glFlush();
	/* contents of back buffer are set to become that of front buffer */
    glutSwapBuffers();
}

/*
=================
keys
-handles all normal keyboard input
==================
*/
void keys(unsigned char key, int x, int y)
{
	/* escape key handling code */
	if (key == ESCAPE)
	{
		exit(0);
	}

	/*
	key handling code for the rotation of our scene
	*/
	if (key == 'a')
		theta += 5;

	if (key == 'd')
		theta -= 5;

	/* sets the current window for redisplay */
	glutPostRedisplay();
}

/*
=================
myInit
-handles basic initialisation and setup of the OpenGL parameters
=================
*/
void myInit()
{
	/* enable texturing */
	glEnable(GL_TEXTURE_2D);
	/* enable hidden-surface removal */
	glEnable(GL_DEPTH_TEST);
	/* specify the value used for depth buffer comparisons
	   - GL_LEQUAL: if incoming z value is <= to the stored z value */
	glDepthFunc(GL_LEQUAL);
	textureLoader();
}

/*
================
reshape
-Rendered window's reshape function - manages changes to size of the window, correctly adjusting aspect ratio.
=================
*/
void reshape(int w, int h)
{
	/* sets the viewport with the first two parameters the lower-left corner of the viewport 
	   rectangle, in pixels and the last two, the width and height of the viewport */
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);

	/* specifies the Projection matrix as the current matrix */
	glMatrixMode(GL_PROJECTION);
	/* move to the center of the screen */
	glLoadIdentity();

	/* sets up the  perspective projection matrix:
	   - the first parameter specifies the field of view (fov), an angle in degrees 
	     (in the y-direction)
	   - the second parameter is aspect ratio determining fov 
		 (in the x-direction)
	   - the third parameter is the distance from the viewer to the near clipping plane
	   - the last parameter is the distance from the viewer to the far clipping plane */
	gluPerspective(45, (GLdouble)w / (GLdouble)h, 0.05, 100);
	/* specifies the Modelview matrix as the current matrix */
	glMatrixMode(GL_MODELVIEW);
}

/*
================
getRandomDouble
-generic function that returns a random double value between the provided bounds
-used with the setup of our heightmap and the animations
================
*/
double getRandomDouble(double Low, double High)
{
	return ((double) rand() / (RAND_MAX+1.0)) * (High-Low) + Low;
}

/*
=================
setupHeights
-creates a randomly perturbed fluid surface, and a more structured rock/ground surface.
=================
*/
void setupHeights()
{
	for (int i = 0; i < g_mapDimension; i++)
	{
		for (int j = 0; j < g_mapDimension; j++)
		{
			wireframeHeightMap[i][j] = getRandomDouble(0.0, 0.1);
			surfaceHeightMap[i][j] = sin(0.25*(sqrt((double)(i)*(double)(i) + ((double)j)*(double)(j)))) / 4.0 - 0.075;
		}
	}
}

/*
=================
idle
-used for animation of our scene
=================
*/
void idle()
{
	/* the mesh animation speed */
	duration = 50;

	if(duration <= duration_counter)
	{
		for (int y = 0; y < g_mapDimension; y++)
		{
			for(int x = 0; x < g_mapDimension; x++)
			{
				temp_map[x][y] = getRandomDouble(0.0, 0.1);
			}
		}
		duration_counter = 0;
	}
	else
	{
		for (int y = 0; y < g_mapDimension; y++)
		{
			for(int x = 0; x < g_mapDimension; x++)
			{
				wireframeHeightMap[x][y] += (temp_map[x][y] - wireframeHeightMap[x][y])/(duration+0.0);
			}
		}
		duration_counter++;
	}
	glutPostRedisplay();
}

/*
=================
main
=================
*/
int main(int argc, char **argv)
{
	/* initialize the GLUT library. */
    glutInit(&argc, argv);
	/* sets the initial display mode: double buffering and a depth buffer. */
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	/* sets default window size */
    glutInitWindowSize(g_windowWidth, window_height);
	/* creates a top-level window with a caption */
    glutCreateWindow("Two height maps - the one a wire frame mesh, the other a surface");
	/* reshape callback for the current window */
    glutReshapeFunc(reshape);
	/* displays our contents in the window */
    glutDisplayFunc(display);
	/* keyboard handling */
    glutKeyboardFunc(keys);
	/* sets the global idle callback for background processing/animation */
    glutIdleFunc(idle);
	myInit();
	setupHeights();
	/* enters the GLUT event processing loop - will handle any callback */
	glutMainLoop();
}
//EOF